//  supporting information of "Inherent Atomic Mobility Changes in Carbocation Intermediates during the Sesterterpene Cyclization Cascade”,
//  in the Beilstein Journal of Organic Chemistry
//

#include <iostream>
#include <string>
#include <fstream>
#include <cstring>
#include <cstdio>
#include <cstdlib>
#include <iomanip>
#include <math.h>

#define PI 3.14159265358979323846264338327950288419716939937510

//prototype declaration
void convert(double *x, double *y, double *z, double parameter[][8], int j);  //coordinate conversion
long double rotA(double x, double y, double radx); //rotation
long double rotB(double x, double y, double radx); //rotation
void trajectory(double x, double y, double z, double **sum, double **derivatives, int atom_number, int step_number); // summing trajectory


using namespace std;

int main (int argc, char *argv[]){
    
    if(argc == 1){
        cout << "\nERROR -- no argumanet --\nirc [file_name]\n" << endl;
        exit(1);
    }

//chose 3 carbon atom numbers
    int carbon_A(0), carbon_B(0), carbon_C(0);
    cout << "Please input three carbon number: ";
    cin >> carbon_A >> carbon_B >> carbon_C;
    
    double **A_coordinate = new double *[3000];
    double **B_coordinate = new double *[3000];
    double **C_coordinate = new double *[3000];
    
    for(int i(0);i<3000;i++){
        A_coordinate[i] = new double[3];
        B_coordinate[i] = new double[3];
        C_coordinate[i] = new double[3];
    }

//serching the first and last line of the coordinate

    long *start_line;
    start_line = new long[3000];
    long *end_line;
    end_line = new long[3000];
    
    ifstream ifs(argv[1]);
    string line;
    long i(0);
    int k(0); //number_of_coordinates
    
    //put all coordinates into arrays
    
    while (getline(ifs,line)){
        i++;
        
        std::string::size_type start = line.find("Number     Number       Type             X           Y           Z");
        std::string::size_type end = line.substr(0,17).find("SCF Done:  E(scf)");
        if (start != std::string::npos ){
            start_line[k] = i;
        }else if (i == start_line[k] + carbon_A + 1){
            int a, b, c;
            double x, y, z;
            sscanf( line.data(), "%d %d %d %lf %lf %lf\n", &a, &b, &c, &x, &y, &z);
            
            A_coordinate[k][0] = x;
            A_coordinate[k][1] = y;
            A_coordinate[k][2] = z;
        }else if (i == start_line[k] + carbon_B + 1){
            int a, b, c;
            double x, y, z;
            sscanf( line.data(), "%d %d %d %lf %lf %lf\n", &a, &b, &c, &x, &y, &z);
            
            B_coordinate[k][0] = x;
            B_coordinate[k][1] = y;
            B_coordinate[k][2] = z;
        }else if (i == start_line[k] + carbon_C + 1){
            int a, b, c;
            double x, y, z;
            sscanf( line.data(), "%d %d %d %lf %lf %lf\n", &a, &b, &c, &x, &y, &z);
            
            C_coordinate[k][0] = x;
            C_coordinate[k][1] = y;
            C_coordinate[k][2] = z;
        }else if (end != std::string::npos ){
            end_line[k]=i;
            k++;
        }
    }
    
    end_line[k] = start_line[k] + end_line[1] - start_line[1];//This process is because the last coodinate doesn't have "SCF DONE".
    
    int total_steps = k + 1;
    int total_atoms = end_line[1] - start_line[1] - 3;
    cout << total_atoms << endl;
    
    double parameter[total_steps][8];
  
    //the following is the overlaying process
    
    for(int j(0);j<=total_steps;j++){
        parameter[j][0] = A_coordinate[j][0];  //parallel shift of atom_A
        parameter[j][1] = A_coordinate[j][1];  //parallel shift of atom_A
        parameter[j][2] = A_coordinate[j][2];  //parallel shift of atom_A
        parameter[j][3] = atan((B_coordinate[j][0]-A_coordinate[j][0])/(B_coordinate[j][1]-A_coordinate[j][1]));  //rotation of atom_B
        
        double B_temp_x = rotA((B_coordinate[j][0]-A_coordinate[j][0]),(B_coordinate[j][1]-A_coordinate[j][1]),parameter[j][3]);
        double B_temp_y = rotB((B_coordinate[j][0]-A_coordinate[j][0]),(B_coordinate[j][1]-A_coordinate[j][1]),parameter[j][3]);
        double B_temp_z = B_coordinate[j][2]-A_coordinate[j][2];
        
        parameter[j][4] = atan(B_temp_y/B_temp_z);  //rotation of atom_B
        if(rotB(B_temp_y,B_temp_z,parameter[j][4])<0)parameter[j][4] += PI;
        
        double B_temp_yy = B_temp_y;
        B_temp_y = rotA(B_temp_yy,B_temp_z,parameter[j][4]);
        B_temp_z = rotB(B_temp_yy,B_temp_z,parameter[j][4]);
        
        //process for converting C_carbon's conversion and setting parameter[j][5].
        //1.parallel shift & 2. rotation
        double C_temp_x = rotA((C_coordinate[j][0]-A_coordinate[j][0]),(C_coordinate[j][1]-A_coordinate[j][1]),parameter[j][3]);
        double C_temp_y = rotB((C_coordinate[j][0]-A_coordinate[j][0]),(C_coordinate[j][1]-A_coordinate[j][1]),parameter[j][3]);
        double C_temp_z = C_coordinate[j][2]-A_coordinate[j][2];
        //3.rotation
        //double temp_radian = atan(C_temp_y/C_temp_z);
        double C_temp_yy = C_temp_y;
        C_temp_y = rotA(C_temp_y,C_temp_z,parameter[j][4]);
        C_temp_z = rotB(C_temp_yy,C_temp_z,parameter[j][4]);
        
        parameter[j][5] = atan(C_temp_x/C_temp_y);  //rotation of atom_C
        if(rotB(C_temp_x,C_temp_y,parameter[j][5])<0)parameter[j][5] += PI;
        
        C_temp_y = rotB(C_temp_x,C_temp_y,parameter[j][5]);
        B_temp_y = rotB(B_temp_x,B_temp_y,parameter[j][5]);
        //center of gravity parameter
        
        parameter[j][6] = (B_temp_y + C_temp_y)/3;
        parameter[j][7] = (B_temp_z + C_temp_z)/3;
        
    }
    
    //deleting memory
    
    for(int i(0);i<3000;i++){
        delete[] A_coordinate[i];
        delete[] B_coordinate[i];
        delete[] C_coordinate[i];
        A_coordinate[i] = 0;
        B_coordinate[i] = 0;
        C_coordinate[i] = 0;
    }
    
    delete[] A_coordinate;
    delete[] B_coordinate;
    delete[] C_coordinate;
    A_coordinate = 0;
    B_coordinate = 0;
    C_coordinate = 0;
    

    ifs.clear();
    ifs.seekg(0, ios_base::beg);
    
    //Output File
    ofstream ofs("output.log",ios::app);
    
    i = 0;
    int j(0);
    while (getline(ifs,line)){
        i++;
        
        if(i > start_line[j]+1 && i < end_line[j] -1){
            int a, b, c;
            double x, y, z;
            sscanf( line.data(), "%d %d %d %lf %lf %lf\n", &a, &b, &c, &x, &y, &z);
            
            ofs<< a << "\t" << b << "\t" << c << "\t";
            
            convert(&x, &y, &z, parameter, j);
            
            if(x<0) ofs<<"  "<<fixed<<setprecision(8)<<x;
            else ofs<<"   "<<fixed<<setprecision(8)<<x;
                
            if(y<0) ofs<<"  "<<fixed<<setprecision(8)<<y;
            else ofs<<"   "<<fixed<<setprecision(8)<<y;
                
            if(z<0) ofs<<"  "<<fixed<<setprecision(8)<<z<<endl;
            else ofs<<"   "<<fixed<<setprecision(8)<<z<<endl;
         
        }else if(i == end_line[j]){
            j++;
            ofs << line << endl;
        }else ofs << line << endl;
    }
    cout << "Normal Termination" << endl;
    
    ifs.close();
    ofs.close();
    
    double **sum = new double *[total_atoms+2];
    sum[0] = new double[(total_atoms+2)*5];
    for(int h(0);h<total_atoms+2;h++)sum[h] = sum[0] + h*5;
    
    double **derivatives = new double *[total_atoms+1];
    derivatives[0] = new double[(total_steps+1)*(total_atoms+1)];
    for(int h(0);h<total_steps+1;h++)derivatives[h] = derivatives[0] + h*total_atoms;
    
    ifstream ifs2("output.log");
    i = 0;
    j =0;
    while (getline(ifs2,line)){
        i++;
        
        if(i > start_line[j]+1 && i < end_line[j] - 1){
            int a, b, c;
            double x, y, z;
            sscanf( line.data(), "%d %d %d %lf %lf %lf\n", &a, &b, &c, &x, &y, &z);
            //printf("%lf\t %lf\t %lf\n",x, y, z);
            sum[a][4] = b;
            trajectory(x, y, z, sum, derivatives, a, j);
            
        }else if(i == end_line[j]){
            j++;
        }else continue;
    }
    
    for(j=1;j<total_atoms+1;j++){
        if(sum[j][4] != 1){
            cout << "C";
            if(j < 10) cout << " ";
            cout << j << " trajectory = " << sum[j][3] << endl;
        }
    }
    
    delete[] start_line;
    delete[] end_line;
    start_line = 0;
    end_line = 0;
    
    
    ifs2.close();
    
    ofstream ofs2("output_all.csv",ios::app);
    
    for(j=1;j<total_atoms+1;j++){
        for(int h=1;h<total_steps+1;h++){
            if(h==1) ofs2 << "atom_number: " << j << ",\t";
            ofs2 << fixed << setprecision(8) << derivatives[h][j] << ",";
            if(h==total_steps) ofs2 << endl;
        }
    }
    ofs2.close();
    
    ofstream ofs3("output_carbon.csv",ios::app);
    
    for(j=1;j<total_atoms+1;j++){
        for(int h=1;h<total_steps+1;h++){
            if(sum[j][4] != 1){
                if(h==1) ofs3 << "atom_number: " << j << ",\t";
                ofs3 << fixed << setprecision(8) << derivatives[h][j] << ",";
                if(h==total_steps) ofs3 << endl;
            }else continue;
        }
    }
    
    ofs3.close();
    
}
//end of main function


//function for coordinate conversion

void convert(double *x, double *y, double *z, double parameter[][8], int j){
    //1.Parallel Shift
    double converted_x(0);
    double converted_y(0);
    double converted_z(0);
    double temp_x = *x - parameter[j][0];
    double temp_y = *y - parameter[j][1];
    double temp_z = *z - parameter[j][2];
    
    //2.rotation around z axis
    converted_x = rotA(temp_x,temp_y,parameter[j][3]);
    converted_y = rotB(temp_x,temp_y,parameter[j][3]);
    
    //3.rotation around x axis
    temp_y = converted_y;
    converted_y = rotA(temp_y,temp_z,parameter[j][4]);
    converted_z = rotB(temp_y,temp_z,parameter[j][4]);
    
    //4.rotation around z axis
    temp_x = converted_x;
    temp_y = converted_y;
    converted_x = rotA(temp_x,temp_y,parameter[j][5]);
    converted_y = rotB(temp_x,temp_y,parameter[j][5]);
    
    *x = converted_x;
    *y = converted_y - parameter[j][6];
    *z = converted_z - parameter[j][7];

}


//function for rotation

long double rotA(double x, double y, double radx){
    double A = x * cos( radx ) - y * sin( radx );
    return A;
}
//////////////////////////////////////////////////////////////////////////////////////////
long double rotB(double x, double y, double radx){
    double B = x * sin( radx ) + y * cos( radx );
    return B;
}


void trajectory(double x, double y, double z, double **sum, double **derivatives, int atom_number, int step_number){

    derivatives[step_number][atom_number] = sqrt(pow((sum[atom_number][0]-x),2) + pow((sum[atom_number][1]-y),2) + pow((sum[atom_number][2]-z),2));
    sum[atom_number][3] += derivatives[step_number][atom_number];

    sum[atom_number][0] = x;
    sum[atom_number][1] = y;
    sum[atom_number][2] = z;
    
    
}

